{% load i18n %}

<table class="table table-striped table-width-fix" id="{{ table_name }}">
    <thead>
        <tr>
            <th>{% translate 'Name' %}</th>
            <th>{% translate 'Qty' %}</th>
            <th>{% translate 'Type' %}</th>
            <th>{% translate 'Category' %}</th>
            <th>{% translate 'Price' %}</th>
            <th>{% translate 'Total' %}</th>
        </tr>
    </thead>

    <tbody></tbody>

    <tfoot>
        <tr>
            <th></th>
            <th></th>
            <th></th>
            <th></th>
            <th>{% translate 'Grand Total:' %}</th>
            <th></th>
        </tr>
    </tfoot>

</table>
