# Generated by Django 4.0.9 on 2023-02-16 19:20

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("memberaudit", "0003_add_notify_permission"),
    ]

    operations = [
        migrations.AddField(
            model_name="character",
            name="is_disabled",
            field=models.BooleanField(
                default=False,
                help_text="Disabled characters are no longer updated from ESI.",
            ),
        ),
    ]
